#------------------------------------------------------------------------------
#
#   Commandophone - Engine
#
#------------------------------------------------------------------------------

import itertools
from pygame import draw

#------------------------------------------------------------------------------

radius_per_semitone = 10
pixels_per_scale_step = 10
grid_size = 10


default_instruments = [
	0, 6, 13, 14, 19, 24, 40, 56
]

#------------------------------------------------------------------------------

major_scale = [0, 2, 4, 5, 7, 9, 11]
minor_scale = [0, 1, 3, 5, 7, 9, 10]

scales = [major_scale, minor_scale]
scale_names = ["Major", "Minor"]

note_names = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]

#------------------------------------------------------------------------------

class Voice(object):

	def __init__(self, number, device):
		self.number = number
		self.device = device
		self.channel = number
		self.octave = 0
		self.volume = 127
		self.set_instrument(default_instruments[number])
	
	def get_instrument(self):
		return self.instrument_no
	
	def set_instrument(self, instrument_no):
		self.instrument_no = instrument_no
		self.device.set_instrument(instrument_no, self.channel)
	
	instrument = property(get_instrument, set_instrument)

#	def start_note(self, radius):
#		pitch = self.radius_to_pitch(radius)
#		volume = self.volume
#		channel = self.channel
#		#print "start_note:", pitch, volume, channel ###
#		self.device.note_on(pitch, volume, channel)
#		return (pitch, channel)
	
	def start_note(self, pitch):
		volume = self.volume
		channel = self.channel
		#print "start_note:", pitch, volume, channel ###
		self.device.note_on(pitch, volume, channel)
		return (pitch, channel)
	
#	def radius_to_pitch(self, radius):
#		return 12 * (self.octave + 1) - int(round(radius / radius_per_semitone))

#------------------------------------------------------------------------------

class Circle(object):
	active_note = None
	
	numbers = itertools.count(1)

	def __init__(self, center, voice_no = 0):
		self.number = self.numbers.next()
		self.center = center
		self.voice_no = voice_no
		self.radius = 5
		self.motion = 1
	
	def __str__(self):
		return "Circle(%s)" % self.number
	
	def contains_point(self, point):
		return distsq(self.center, point) < self.radius ** 2
	
	def update_radius(self):
		#print "Circle.update:", self, "radius =", self.radius, "motion =", self.motion ###
		r = self.radius
		m = self.motion
		if r <= 0 and m < 0:
			m = -m
			self.motion = m
		self.radius = r + m
	
	def check_for_wall_collision(self, rect):
		if self.motion > 0:
			x, y = self.center
			r = self.radius
			if x - r <= rect.left or x + r >= rect.right \
				or y - r <= rect.top or y + r >= rect.bottom:
					self.reverse_motion()
					return True
		return False
	
	def reverse_motion(self):
		self.motion *= -1
	
	def trigger(self, engine):
		self.silence(engine)
		self.active_note = engine.start_note(self.voice_no, self.radius)
	
	def silence(self, engine):
		engine.stop_note(self.active_note)

#------------------------------------------------------------------------------

class Engine(object):
	#  bounds     Rect
	#  device     pygame.midi.OutputDevice
	#  circles    [Circle]
	#  voices     [Voice]
	#  scale_tonic   int
	#  scale_mode    int
	#  circle_containment   set of (inner Circle, outer circle)
	#  grid_enabled         boolean
	#  show_centers         boolean
	#  selected_voice_no    int
	#  selected_circle      Circle or None
	
	def __init__(self, device, num_voices):
		self.device = device
		self.circles = []
		self.voices = [Voice(i, device) for i in xrange(num_voices)]
		self.scale_tonic = 0
		self.scale_mode = 0
		self.selected_voice_no = 0
		self.circle_containment = set()
		self.grid_enabled = True
		self.show_centers = True
		self.selected_circle = None
	
	selected_voice = property(lambda self: self.voices[self.selected_voice_no])
	
	def add_circle(self, center):
		center = self.snap_to_grid(center)
		new_circle = Circle(center, self.selected_voice_no)
		self.insert_circle(new_circle)
	
	def insert_circle(self, new_circle):
		center = new_circle.center
		for old_circle in self.circles:
			if old_circle.contains_point(center):
				self.circle_containment.add((new_circle, old_circle))
		self.circles.append(new_circle)
	
	def delete_circle(self, circle):
		circle.silence(self)
		self.remove_circle(circle)
	
	def remove_circle(self, circle):
		circles = self.circles
		if circle in circles:
			circles.remove(circle)
		cont = self.circle_containment
		for pair in list(cont):
			if circle in pair:
				cont.remove(pair)
	
	def move_circle(self, circle, center):
		new_center = self.snap_to_grid(center)
		if new_center <> circle.center:
			self.remove_circle(circle)
			circle.center = new_center
			self.insert_circle(circle)
	
	def snap_to_grid(self, p):
		if self.grid_enabled:
			x, y = p
			g = grid_size
			x = g * (x // g)
			y = g * (y // g)
			p = x, y
		return p

	def update(self):
		circles = self.circles
		for circle in circles:
			circle.update_radius()
		r = self.bounds
		for circle in circles:
			if circle.check_for_wall_collision(r):
				#print "Engine.update:", circle, "collided with wall"
				circle.trigger(self)
		n = len(circles)
		for i in xrange(n):
			circle1 = circles[i]
			for j in xrange(i+1, n):
				circle2 = circles[j]
				if self.circle_inside(circle1, circle2):
					collided = interior_collision_check(circle1, circle2)
				elif self.circle_inside(circle2, circle1):
					collided = interior_collision_check(circle2, circle1)
				else:
					collided = exterior_collision_check(circle1, circle2)
				if collided:
					circle1.reverse_motion()
					circle2.reverse_motion()
					circle1.trigger(self)
					circle2.trigger(self)

	def circle_inside(self, circle1, circle2):
		return (circle1, circle2) in self.circle_containment
	
#	def start_note(self, voice_no, radius):
#		return self.voices[voice_no].start_note(radius)
	
	def start_note(self, voice_no, radius):
		voice = self.voices[voice_no]
		pitch = self.radius_to_pitch(radius, voice.octave)
		return voice.start_note(pitch)
	
	def radius_to_pitch(self, radius, octave_shift):
		k = int(round(radius / pixels_per_scale_step))
		scale = scales[self.scale_mode]
		n = len(scale)
		i = n - 1 - k + self.scale_tonic
		octave, note = divmod(i, n)
		pitch = scale[note] + 12 * (5 + octave + octave_shift)
		return pitch
		
	def stop_note(self, note):
		if note:
			pitch, channel = note
			self.device.note_off(pitch, 0, channel)

#------------------------------------------------------------------------------

def distsq(p1, p2):
	x1, y1 = p1
	x2, y2 = p2
	return (x1 - x2) ** 2 + (y1 - y2) ** 2

def exterior_collision_check(circle1, circle2):
	if circle1.motion > 0 and circle2.motion > 0:
		d1 = distsq(circle1.center, circle2.center)
		d2 = (circle1.radius + circle2.radius) ** 2
		return d2 >= d1
	else:
		return False

def interior_collision_check(inner_circle, outer_circle):
	if inner_circle.motion > outer_circle.motion:
		d1 = distsq(inner_circle.center, outer_circle.center)
		d2 = (outer_circle.radius - inner_circle.radius) ** 2
		return d1 >= d2
	else:
		return False


