#------------------------------------------------------------------------------
#
#   Commandophone - Main
#
#------------------------------------------------------------------------------

from __future__ import division
import sys
import pygame.font
pygame.font.init()
from pygame import Rect, Color
from pygame import draw, font, midi #, gfxdraw as gfx
from pygame.display import set_mode
from pygame.locals import *
from albow.widget import Widget
from albow.root import RootWidget, get_root
from albow.layout import Row, Column, Frame
from albow.references import Ref
from albow.controls import Control, Label, ValueDisplay, Button
from albow.fields import IntField
from albow.dialogs import Dialog
from albow.resource import get_font
#from albow.multichoice import TextMultichoice
from engine import Engine, distsq, note_names, scale_names
from instruments import instrument_names
from controls import Slider, SpinArrows, IntDisplay, ListBox
import midi
import theme

screen_size = (800, 600)
frame_time = 50
midi_device_name = "SimpleSynth virtual input"
center_display_size = 3

colors = [
	(0xff, 0, 0), (0xff, 0x80, 0), (0xff, 0xff, 0),
	(0, 0xff, 0), (0, 0xff, 0xff), (0, 0, 0xff),
	(0x80, 0, 0xff), (0xff, 0xff, 0xff)
]

#------------------------------------------------------------------------------

class InstrumentField(ValueDisplay):

	border_width = 1
	margin = 3

	def format_value(self, value):
		try:
			return instrument_names[value]
		except IndexError:
			return str(value)
	
	def mouse_down(self, e):
		dlog = InstrumentDialog(ref = self.ref)
		dlog.topleft = self.local_to_global((self.width, -200))
		self.get_root().add(dlog)
		dlog.focus()

#------------------------------------------------------------------------------

class NoteDisplay(ValueDisplay):

	border_width = 1
	margin = 3

	def __init__(self, width = "C#", **kwds):
		ValueDisplay.__init__(self, width = width, **kwds)

	def format_value(self, value):
		return note_names[value]

#------------------------------------------------------------------------------

class ModeDisplay(ValueDisplay):

	border_width = 1
	margin = 3

	def __init__(self, width = "Major", **kwds):
		ValueDisplay.__init__(self, width = width, **kwds)

	def format_value(self, value):
		return scale_names[value]

#------------------------------------------------------------------------------

class InstrumentSelector(ListBox):

	font = get_font(10, "Vera.ttf")

	def __init__(self, **kwds):
		ListBox.__init__(self, labels = instrument_names,
			nrows = 32, ncols = 4, cell_width = 150, **kwds)
		
#------------------------------------------------------------------------------

class InstrumentDialog(Dialog):

	border_width = 0
	button_font = get_font(10, "Vera.ttf")

	def __init__(self, ref, **kwds):
		Dialog.__init__(self, **kwds)
		box = InstrumentSelector(ref = ref)
		close_button = Button("Close", font = self.button_font, action = self.cancel)
		contents = Column([box, close_button], align = 'r')
		contents.topleft = (5, 5)
		self.add(contents)
		self.shrink_wrap()
	
	def dismiss(self, value):
		self.get_root().remove(self)

#------------------------------------------------------------------------------

class CircleView(Widget):

	def __init__(self, width, height, engine):
		bounds = Rect(0, 0, width, height)
		engine.bounds = bounds
		Widget.__init__(self, bounds)
		self.engine = engine
	
	def mouse_down(self, e):
		#print "CircleView.mouse_down:", e.local ###
		self.focus()
		p = e.local
		circle = self.find_circle(p)
		self.engine.selected_circle = circle
		if not circle:
			self.engine.add_circle(e.local)
	
	def mouse_drag(self, e):
		p = e.local
		if self.engine.bounds.collidepoint(p):
			circle = self.engine.selected_circle
			if circle:
				self.engine.move_circle(circle, p)
		
	def find_circle(self, p):
		d = center_display_size ** 2
		for circle in self.engine.circles:
			if distsq(circle.center, p) <= d:
				return circle
	
	def key_down(self, e):
		k = e.key
		if k == K_BACKSPACE or k == K_DELETE:
			circle = self.engine.selected_circle
			if circle:
				self.engine.delete_circle(circle)
		else:
			Widget.key_down(self, e)
			
	def draw(self, surf):
		engine = self.engine
		for circle in engine.circles:
			if engine.show_centers:
				self.draw_center(surf, circle)
			self.draw_circle(surf, circle)
	
	def draw_circle(self, surf, circle):
		r = circle.radius
		if r >= 1:
			draw.circle(surf, colors[circle.voice_no], circle.center, r, 1)
		
	def draw_center(self, surf, circle):
		d = center_display_size
		c = colors[circle.voice_no]
		x, y = circle.center
		if circle is self.engine.selected_circle:
			w = 3
		else:
			w = 1
		line = draw.line
		line(surf, c, (x-d, y), (x+d, y), w)
		line(surf, c, (x, y-d), (x, y+d), w)

#------------------------------------------------------------------------------

class VoiceSwatch(Widget):

	selection_color = (0, 0, 0)

	def __init__(self, engine, voice_no):
		self.engine = engine
		self.voice_no = voice_no
		Widget.__init__(self, Rect(0, 0, 30, 20))
	
	def draw(self, surf):
		surf.fill(colors[self.voice_no])
		if self.is_selected():
			r = surf.get_rect()
			x1 = r.centerx - 5
			x2 = x1 + 10
			y1 = r.top + 5
			y2 = r.centery
			y3 = r.bottom - 5
			draw.polygon(surf, self.selection_color,
				[(x1, y1), (x2, y2), (x1, y3)])
	
	def is_selected(self):
		return self.engine.selected_voice_no == self.voice_no
	
	def mouse_down(self, e):
		self.engine.selected_voice_no = self.voice_no

#------------------------------------------------------------------------------

class VoiceView(Row):

	def __init__(self, engine, voice_no):
		self.engine = engine
		self.voice_no = voice_no
		voice = engine.voices[voice_no]
		voice_ref = Ref(voice)
		swatch = VoiceSwatch(engine, voice_no)
		volume_ctrl = IntField(ref = voice_ref.volume,
			min = 0, max = 127)
		program_ctrl = IntField(ref = voice_ref.instrument,
			min = 0, max = 127)
		octave_ctrl = IntField(ref = voice_ref.octave,
			min = -5, max = 5)
		Row.__init__(self, [swatch, volume_ctrl, program_ctrl, octave_ctrl],
			spacing = 2)

#------------------------------------------------------------------------------

class SelectedVoiceFrame(Frame):

	border_width = 0
	margin = 5

	def __init__(self, client, engine):
		Frame.__init__(self, client)
		self.engine = engine

	def draw(self, surf):
		surf.fill(colors[self.engine.selected_voice_no])

#------------------------------------------------------------------------------

class CPRoot(RootWidget):

	def setup(self, engine):
		self.engine = engine
		controls = self.create_controls(engine)
		view = CircleView(self.width - controls.width, self.height, engine)
		view.left = controls.right
		self.add(controls)
		self.add(view)
		view.focus()

	def create_controls(self, engine):
		ref_engine = Ref(engine)
		voice_ctrls = Column([VoiceView(engine, voice_no)
			for voice_no in xrange(len(engine.voices))],
			spacing = 2)
		volume_ctrl = Slider(width = 150, height = 20,
			min = 0, max = 127, ref = ref_engine.selected_voice.volume)
		instrument_ctrl = InstrumentField(width = 150,
			ref = ref_engine.selected_voice.instrument)
		ref_octave = ref_engine.selected_voice.octave
		octave_view = IntDisplay(width = "-5", ref = ref_octave)
		octave_spin = SpinArrows(min = -5, max = 5, ref = ref_octave)
		current_voice_ctrls = SelectedVoiceFrame(
			engine = engine,
			client = Column([
				Label("Volume"),
				volume_ctrl,
				Label("Instrument"),
				instrument_ctrl,
				Row([Label("Octave"), octave_view, octave_spin], spacing = 5),
			], spacing = 2, align = 'l'))
		ref_key = ref_engine.scale_tonic
		key_view = NoteDisplay(ref = ref_key)
		key_spin = SpinArrows(min = 0, max = 11, wrap = True, ref = ref_key)
		ref_mode = ref_engine.scale_mode
		mode_view = ModeDisplay(ref = ref_mode)
		mode_spin = SpinArrows(min = 0, max = 1, wrap = True, ref = ref_mode)
		controls = Column([
			voice_ctrls,
			#Label("Selected Voice"),
			current_voice_ctrls,
			#Label("Global Options"),
			Row([Label("Key"), key_view, key_spin, mode_view, mode_spin], spacing = 5),
		], align = 'l', margin = 5)
		controls.height = self.height
		controls.bg_color = (0x80, 0x80, 0x80)
		return controls
	
	def timer_event(self, e):
		self.engine.update()
		return True

	def key_down(self, e):
		k = e.key
		if k == K_q and e.cmd:
			self.quit()
		else:
			RootWidget.key_down(self, e)

#------------------------------------------------------------------------------

def main():
	display = set_mode(screen_size)
	root = CPRoot(display)
	#device = init_midi()
	device = midi.open_output_device()
	if not device:
		return
	try:
		engine = Engine(device, len(colors))
		root.setup(engine)
		root.set_timer(frame_time)
		root.run()
	finally:
		midi.all_sound_off(device)
